from pydantic.fields import FieldInfo

from .types import Dependant, Field

from typing import TYPE_CHECKING

import inspect

if TYPE_CHECKING:
    from typing import Callable


def get_param_field(name: str, param: inspect.Parameter) -> Field:
    if not param.annotation:
        raise ValueError(f'The parameter "{name}" does not have a type annotation')

    kwargs = {'annotation': param.annotation}

    if param.default != param.empty:
        kwargs['default'] = param.default

    field_info = FieldInfo(**kwargs)

    return Field(
        name=name,
        field_info=field_info,
    )


def get_dependant(*, handler: 'Callable') -> Dependant:
    signature = inspect.signature(handler)
    signature_params = signature.parameters

    dependant = Dependant()

    for name, param in signature_params.items():
        field = get_param_field(name, param)
        dependant.params.append(field)

    return dependant


def none_param_sort(key: str) -> callable:
    def inner(element: any) -> None:
        return getattr(element, key) or ''

    return inner



# In this example, we are sending messages so that the code from the first example can process them

from aiocarrot import Carrot


async def main() -> None:
    """ Entrypoint of example application """

    carrot = Carrot(url='amqp://guest:guest@localhost:5672/', queue_name='users')

    await carrot.send('register_user', user_id=1)
    await carrot.send('add_points', user_id=1, amount=1000)
    await carrot.send('add_points', user_id=1)

    # If you look at the example 01_consumer.py then you will see that the amount argument
    # is optional and by default the user is awarded 100 points

if __name__ == '__main__':
    import asyncio

    asyncio.run(main())



from aiocarrot import Carrot, Consumer

from contextlib import asynccontextmanager

import pytest, asyncio


@asynccontextmanager
async def get_test_client(carrot_configuration: dict[str, str]) -> Carrot:
    """ Returns the Carrot test client """

    carrot = Carrot(**carrot_configuration)

    async def run_carrot_loop() -> None:
        nonlocal carrot

        carrot.setup_consumer(Consumer())

        await carrot.run()

    task = asyncio.create_task(run_carrot_loop())

    yield carrot

    task.cancel()

    while not task.done():
        await asyncio.sleep(0.1)


@pytest.fixture
def carrot_configuration() -> dict[str, str]:
    """ Standard configuration for connecting to RabbitMQ """

    return {
        'url': 'amqp://guest:guest@localhost:5672/',
        'queue_name': 'aiocarrot/test',
    }


@pytest.fixture
async def carrot_client(carrot_configuration: dict[str, str]) -> Carrot:
    """ Carrot test client for working with RabbitMQ """

    async with get_test_client(carrot_configuration) as carrot:
        yield carrot



from pydantic import BaseModel

from loguru import logger

from typing import Optional, TYPE_CHECKING

import asyncio, aio_pika, ujson, uuid, copy, signal

if TYPE_CHECKING:
    from aiormq.abc import ConfirmationFrameType

    from .consumer import Consumer


class Carrot:
    """ Carrot framework entrypoint class """

    _url: str
    _queue_name: str
    _tasks: list[asyncio.Task]
    _is_consumer_alive: bool = False
    _consumer: Optional['Consumer'] = None
    _connection: Optional['aio_pika.abc.AbstractConnection'] = None
    _channel: Optional['aio_pika.abc.AbstractChannel'] = None
    _queue: Optional['aio_pika.abc.AbstractQueue'] = None

    def __init__(self, url: str, queue_name: str) -> None:
        """
        aiocarrot is an asynchronous framework for working with the RabbitMQ message broker

        :param url: RabbitMQ connection url
        :param queue_name: The name of the queue for further work
        """

        self._url = url
        self._tasks = []
        self._queue_name = queue_name

    async def send(self, _cnm: str, **kwargs) -> 'ConfirmationFrameType':
        """
        Send a message with the specified type and the specified payload

        :param _cnm: The name of the message (used to determine the type of message being sent)
        :param kwargs: The payload transmitted in the message body
        :return:
        """

        channel = await self._get_channel()

        message_id = str(uuid.uuid4())
        message_body = {
            '_cid': message_id,
            '_cnm': _cnm,
            **kwargs,
        }

        message_body = {
            key: (value.model_dump() if isinstance(value, BaseModel) else value)
            for key, value in message_body.items()
        }

        payload = ujson.dumps(message_body).encode()

        return await channel.default_exchange.publish(
            message=aio_pika.Message(body=payload),
            routing_key=self._queue_name,
        )

    def setup_consumer(self, consumer: 'Consumer') -> None:
        """
        Sets the consumer as the primary one for this Carrot instance

        :param consumer: Consumer object
        :return:
        """

        self._consumer = consumer

    async def run(self) -> None:
        """
        Starts the main loop of the Carrot new message listener

        :return:
        """

        if not self._consumer:
            raise RuntimeError('Consumer is not registered. Please, specify using following method: '
                               '.setup_consumer(consumer)')

        if self._is_consumer_alive:
            raise RuntimeError('Consumer loop is already running')

        logger.info('Starting aiocarrot with following configuration:')
        logger.info('')
        logger.info(f'> Queue: {self._queue_name}')
        logger.info(f'> Registered messages:')

        for message_name in self._consumer._messages.keys():
            logger.info(f'  * {message_name}')

        logger.info('')
        logger.info('Starting listener loop...')

        signal.signal(signal.SIGINT, self._exit_signal_handler)

        self._is_consumer_alive = True
        consumer_task = asyncio.create_task(self._consumer_loop())

        try:
            while self._is_consumer_alive:
                if consumer_task.done():
                    logger.critical(
                        f'An unhandled error occurred while the consumer was working: '
                        f'{str(consumer_task.exception())}'
                    )
                    break

                await asyncio.sleep(1)
        except asyncio.CancelledError:
            logger.info('Shutdown signal received')

        if not consumer_task.done():
            consumer_task.cancel()

        await self._shutdown()

    async def _consumer_loop(self) -> None:
        """
        The main event loop used by Carrot to receive new messages and pass them on to the handler

        :return:
        """

        queue = await self._get_queue()

        logger.info('Consumer is successfully connected to queue')

        async with queue.iterator() as queue_iterator:
            async for message in queue_iterator:
                for task in copy.copy(self._tasks):
                    if task.done():
                        self._tasks.remove(task)

                async with message.process():
                    decoded_message: str = message.body.decode()

                    try:
                        message_payload = ujson.loads(decoded_message)

                        assert isinstance(message_payload, dict)
                    except ujson.JSONDecodeError:
                        logger.error(f'Error receiving the message (failed to receive JSON): {decoded_message}')
                        continue

                    message_id = message_payload.get('_cid')
                    message_name = message_payload.get('_cnm')

                    if not message_id:
                        logger.error(
                            'The message format could not be determined (identifier is missing): '
                            f'{message_payload}'
                        )

                        continue

                    if not message_name:
                        logger.error(
                            'The message format could not be determined (message name is missing): '
                            f'{message_payload}'
                        )

                        continue

                    del message_payload['_cid']
                    del message_payload['_cnm']

                    task = asyncio.create_task(self._consumer.on_message(
                        message_id,
                        message_name,
                        **message_payload,
                    ))

                    self._tasks.append(task)

    async def _get_queue(self) -> 'aio_pika.abc.AbstractQueue':
        """
        Retrieves the currently active aiopika queue object

        :return: aiopika queue
        """

        if not self._queue:
            channel = await self._get_channel()
            self._queue = await channel.declare_queue(self._queue_name, durable=True, auto_delete=True)

        return self._queue

    async def _get_channel(self) -> 'aio_pika.abc.AbstractChannel':
        """
        Gets the current active object of the aiopika channel

        :return: aiopika channel
        """

        if not self._channel:
            connection = await self._get_connection()
            self._channel = await connection.channel()

        return self._channel

    async def _get_connection(self) -> 'aio_pika.abc.AbstractConnection':
        """
        Retrieves the object of an active connection with the broker using aiopika

        :return: aiopika broker connection
        """

        if not self._connection:
            self._connection = await aio_pika.connect_robust(url=self._url)

        return self._connection

    async def _shutdown(self) -> None:
        """
        Shutdown carrot application

        :return:
        """

        pending_tasks = [x for x in self._tasks if not x.done()]

        if len(pending_tasks) > 0:
            logger.info(f'Waiting for {len(pending_tasks)} tasks...')
            await asyncio.gather(*pending_tasks)

        try:
            await self._channel.close()
        except:
            pass

        try:
            await self._connection.close()
        except:
            pass

        logger.info('Good bye!')

    def _exit_signal_handler(self, sig, frame) -> int:
        logger.info('Shutdown signal received')

        self._is_consumer_alive = False

        return 0


__all__ = (
    'Carrot',
)



from aiocarrot import Consumer
from aiocarrot.exceptions import MessageExistsError


def test_collision() -> None:
    primary_consumer = Consumer()

    @primary_consumer.message(name='unique')
    async def unique_message() -> None:
        return

    @primary_consumer.message(name='wrong')
    async def wrong_message() -> None:
        return

    test_consumer = Consumer()

    @test_consumer.message(name='wrong')
    async def wrong_duplicate_message() -> None:
        return

    try:
        primary_consumer.include_consumer(test_consumer)
    except MessageExistsError:
        pass
    else:
        assert False, 'Collision not detected.'



from pydantic import TypeAdapter, ValidationError, ConfigDict
from pydantic.fields import FieldInfo
from pydantic_core import PydanticUndefined as Undefined

from typing import Callable, NamedTuple, TypeVar, Annotated
from dataclasses import dataclass


T = TypeVar('T')


@dataclass
class Field:
    name: str
    field_info: FieldInfo

    @property
    def required(self) -> bool:
        return self.field_info.is_required()

    @property
    def default(self) -> any:
        if self.field_info.is_required():
            return Undefined
        return self.field_info.get_default(call_default_factory=True)

    @property
    def type(self) -> any:
        return self.field_info.annotation

    def __post_init__(self) -> None:
        self._type_adapter: TypeAdapter[any] = TypeAdapter(
            Annotated[self.field_info.annotation, self.field_info],
        )

    def validate(self, value: any) -> tuple[T | None, str | None]:
        try:
            return self._type_adapter.validate_python(value, from_attributes=True), None
        except ValidationError as exc:
            return None, exc.errors()


class Dependant:
    params: list[Field]

    def __init__(self, *, params: list[Field] = None):
        self.params = params or []


class Message(NamedTuple):
    name: str | None
    handler: Callable
    dependant: Dependant


__all__ = (
    'Field',
    'Dependant',
    'Message',
)



# This is an example of a consumer whose task is to create new users and add points to their balance
#
# In a real project, you can use your own database instead of a dictionary.
# For example, you can use SQLAlchemy ORM to work with SQL databases.


from aiocarrot import Carrot, Consumer


consumer = Consumer()
users: dict[int, int] = {}


@consumer.message(name='register_user')
async def register_user_message(user_id: int) -> None:
    """
    Example of a message handler for registering a new user

    :param user_id: User identifier (this field is required and must be of the int type)
    :return:
    """

    if user_id in users:
        raise ValueError('This user is already registered')

    users[user_id] = 0


@consumer.message(name='add_points')
async def add_points_message(user_id: int, amount: int = None) -> None:
    """
    Adds the number of points to the user

    :param user_id: User identifier
    :param amount: The number of points to be added
    :return:
    """

    if user_id not in users:
        raise ValueError('This user is not registered')

    if amount is None:
        amount = 100

    users[user_id] += amount


async def main() -> None:
    """ Entrypoint of example application """

    carrot = Carrot(url='amqp://guest:guest@localhost:5672/', queue_name='users')
    carrot.setup_consumer(consumer)

    await carrot.run()


if __name__ == '__main__':
    import asyncio

    asyncio.run(main())



from aiocarrot import Carrot, Consumer

from pydantic import BaseModel

from typing import Optional

import asyncio


async def test_validation(carrot_client: Carrot) -> None:
    success_counts = 0
    consumer = Consumer()

    class SampleModel(BaseModel):
        number: int
        optional_number: Optional[int] = None

    @consumer.message(name='full_model')
    async def full_model_message(model: SampleModel) -> None:
        nonlocal success_counts

        if model.number != 10:
            raise ValueError('<number> must be 10')

        if model.optional_number != 20:
            raise ValueError('<optional_number> must be 20')

        success_counts += 1

    @consumer.message(name='empty_model')
    async def empty_model_message(model: SampleModel) -> None:
        nonlocal success_counts

        if model.number != 12:
            raise ValueError('<number> must be 12')

        if model.optional_number is not None:
            raise ValueError('<optional_number> must be None')

        success_counts += 1

    @consumer.message(name='int_variables')
    async def int_variables_message(number: int, optional_number: Optional[int] = None) -> None:
        nonlocal success_counts

        if number != 15:
            raise ValueError('<number> must be 15')

        if optional_number is not None and optional_number != 0:
            raise ValueError('<optional_number> must be empty or 0')

        success_counts += 1

    @consumer.message(name='boolean_variables')
    async def boolean_variables_message(value: bool) -> None:
        nonlocal success_counts

        if value is True:
            raise ValueError('<value> must be False')

        success_counts += 1

    carrot_client.setup_consumer(consumer)

    await asyncio.sleep(1)

    await carrot_client.send('full_model', model=SampleModel(number=10, optional_number=20))
    await carrot_client.send('empty_model', model=SampleModel(number=12))
    await carrot_client.send('int_variables', number=15, optional_number=0)
    await carrot_client.send('int_variables', number=15, optional_number=None)
    await carrot_client.send('int_variables', number=15)
    await carrot_client.send('boolean_variables', value=False)
    await carrot_client.send('boolean_variables', value=True)
    await carrot_client.send('boolean_variables', value=None)
    await carrot_client.send('boolean_variables', value=254)

    await asyncio.sleep(1)  # Wait for tasks completed

    assert success_counts == 6



from ..exceptions import MessageExistsError

from .types import Message
from .utils import get_dependant

from loguru import logger

from abc import ABC
from typing import TYPE_CHECKING
from copy import deepcopy

if TYPE_CHECKING:
    from typing import Callable


class AbstractConsumer(ABC):
    _messages: dict[str, Message]

    def create_message(self, name: str, handler: 'Callable') -> None:
        """
        Store message handler in the consumer

        :param name: Message name
        :param handler: Handler function
        :return:
        """
        raise NotImplementedError

    def message(self, name: str) -> 'Callable':
        """
        Create a new message handler

        :param name:
        :return:
        """
        raise NotImplementedError

    def on_message(self, _cid: str, _cnm: str, **kwargs) -> None:
        """
        Received message event handler

        :param _cid: Message identifier
        :param _cnm: Message name
        :return:
        """
        raise NotImplementedError


class Consumer(AbstractConsumer):
    _messages: dict[str, Message]

    def __init__(self) -> None:
        self._messages = {}

    def create_message(self, name: str, handler: 'Callable') -> None:
        split_name = name.split()

        if len(split_name) != 1:
            raise AttributeError(f'Route name must be single word (found: {name})')

        message = Message(
            name=name,
            handler=handler,
            dependant=get_dependant(handler=handler),
        )

        self._messages[name] = message

    def message(self, name: str | list[str]) -> 'Callable':
        def decorator(func: 'Callable') -> 'Callable':
            if not isinstance(name, list):
                message_names = [name]
            else:
                message_names = name

            for message_name in message_names:
                self.create_message(name=message_name, handler=func)

            return func

        return decorator

    async def on_message(self, _cid: str, _cnm: str, **kwargs) -> None:
        message = self._messages.get(_cnm)

        if not message:
            logger.error(f'[{_cid}] Received unknown message <{_cnm}>: {kwargs}')
            return

        logger.info(f'[{_cid}] Received message <{_cnm}>: {kwargs}')

        values = {}

        for field in message.dependant.params:
            if field.name not in kwargs:
                if field.required:
                    logger.error(f'[{_cid}] Error processing the message <{_cnm}>: field "{field.name}" is required')
                    return
                else:
                    values[field.name] = deepcopy(field.default)

                continue

            value, errors = field.validate(kwargs[field.name])

            if errors is not None:
                logger.error(
                    f'[{_cid}] Error processing the message <{_cnm}>: field "{field.name}" with value "{kwargs[field.name]}" '
                    f'does not meet the validation conditions - {str(errors)}',
                )

                return

            values[field.name] = value

        try:
            await message.handler(**values)
        except BaseException as e:
            logger.error(f'[{_cid}] Message processing failed <{_cnm}>: {str(e)}')
            return

        logger.info(f'[{_cid}] Message <{_cnm}> processed successfully')

    def include_consumer(self, consumer: 'Consumer') -> None:
        """
        Include messages from the transmitted consumer into the current one

        :param consumer: Consumer object
        :return:
        """

        common_keys = list(set(self._messages.keys()) & set(consumer._messages.keys()))

        if len(common_keys) > 0:
            raise MessageExistsError(f'Error when trying to add a duplicate message: {common_keys[0]}')

        self._messages |= consumer._messages


__all__ = (
    'Consumer',
)



# This example shows an example of the code for sending data to a queue using FastAPI


from fastapi import FastAPI, Body
from aiocarrot import Carrot


app = FastAPI()
carrot = Carrot(url='amqp://guest:guest@localhost:5672/', queue_name='users')


@app.post('/signup')
async def register_user_view(user_id: int = Body(embed=True)):
    """ An example of an API method for creating a new user """

    await carrot.send('register_user', user_id=user_id)

    return {'status': True}


@app.post('/reward')
async def reward_view(user_id: int = Body(embed=True), amount: int = Body(embed=True)):
    """
    An example of an API method for receiving a reward in
    the amount of the number of points passed in the request
    """

    await carrot.send('add_points', user_id=user_id, amount=amount)

    return {'status': True}



from aiocarrot import Carrot, Consumer

import asyncio


async def test_messages(carrot_client: Carrot) -> None:
    counter = 0
    consumer = Consumer()

    @consumer.message(name='increase')
    async def increase_message(value: int = 1) -> None:
        nonlocal counter
        counter += value

    @consumer.message(name='decrease')
    async def decrease_message(value: int = 1) -> None:
        nonlocal counter
        counter -= value

    carrot_client.setup_consumer(consumer)

    await asyncio.sleep(1)

    await carrot_client.send('increase', value=15)
    await carrot_client.send('decrease', value=3)
    await carrot_client.send('increase')
    await carrot_client.send('decrease')

    await asyncio.sleep(1)  # Wait for tasks completed

    assert counter == 12



from aiocarrot import Carrot, Consumer

from pydantic import BaseModel

import asyncio


class CaseModel(BaseModel):
    text: str
    value: int
    flag: bool = False


async def test_models(carrot_client: Carrot) -> None:
    text, value, flag = None, None, None
    consumer = Consumer()

    @consumer.message(name='test_model')
    async def increase_message(model: CaseModel) -> None:
        nonlocal text, value, flag
        text, value, flag = model.text, model.value, model.flag

    carrot_client.setup_consumer(consumer)

    await asyncio.sleep(1)

    await carrot_client.send('test_model', model=CaseModel(text='test', value=12))

    await asyncio.sleep(1)  # Wait for tasks completed

    assert text == 'test'
    assert value == 12
    assert flag is False
